import logging

from gym_jass.envs import SchieberEnv
from schieber.card import from_card_to_index
from schieber.player.random_player import RandomPlayer
from stable_baselines.bench import Monitor
from stable_baselines.common.vec_env import DummyVecEnv, SubprocVecEnv

logger = logging.getLogger(__name__)


def delete_invalid_actions(probabilities, obs_dict):
    """
    Set action probabilities of invalid cards to 0
    :param probabilities:
    :param obs_dict:
    :return:
    """
    logger.info(f"action probabilities of all actions: {probabilities}")
    hand_cards = obs_dict["cards"]
    allowed_cards = RandomPlayer().allowed_cards_with_hand_cards(obs_dict, hand_cards)
    # subtract one because in the observation here we have indices from 0 to 35 and not from 1 to 36
    allowed_indices = [from_card_to_index(card) - 1 for card in allowed_cards]
    for j in range(probabilities.size):
        if j not in allowed_indices:
            probabilities[0][j] = 0
    logger.info(f"action probabilities of valid actions: {probabilities}")
    return probabilities


def get_action_with_highest_probability(probabilities):
    """
    Choose the action which should be taken in the end
    :param probabilities:
    :return:
    """
    return [int(probabilities.argmax())]


def init_env(env_name, log_dir, n_cpu=False, rules_reward=False):
    """
    Initializes the environment by a given name
    :param rules_reward: whether to take the rules reward or the stich_reward
    :param log_dir:
    :param env_name:
    :param n_cpu: set a number to enable multiprocessing. Does not work yet for the gym-jass env!
    :return:
    """
    # env = gym.make(env_name)
    # different way to initialize the environment: possibility to parametrize
    env = SchieberEnv(rules_reward=rules_reward)
    env = Monitor(env, log_dir,
                  allow_early_resets=True)  # The file monitor.csv which this Monitor creates is used in the callback.
    if n_cpu > 1:
        # return SubprocVecEnv([lambda: env for _ in range(n_cpu)])
        return DummyVecEnv(
            [lambda: env for _ in range(n_cpu)])  # TODO not sure if this does not result in threading problems!!!
    else:
        return DummyVecEnv([lambda: env])


def evaluate(env, model, n_games=10):
    """
    Evaluates the learned model.
    TODO in a later steps invalid actions could be removed so that only valid actions can be chosen.
    :param env:
    :param model:
    :param n_games:
    :return:
    """
    for i in range(n_games):
        obs = env.reset()
        for j in range(9):
            action = model.predict(obs)[0]  # the prediction is a tuple with the next state as the second entry
            obs, rewards, dones, info = env.step(action)
            print(f"Reward: {rewards}")
            env.render()



# TODO ask to be mentioned on https://stable-baselines.readthedocs.io/en/master/misc/projects.html
# TODO add uct search
# TODO what do these metrics from baselines mean:
# --------------------------------------
# | approxkl           | 0.00069229305 | 0.1 -> small, 10 -> large
# | clipfrac           | 0.00390625    |
# | ep_rewmean         | nan           | mean episode reward!
# | eplenmean          | nan           | mean episode length
# | explained_variance | 0.847         | when 0 -> tune hyper-parameters of neural network
# | fps                | 142           |
# | nupdates           | 782           |
# | policy_entropy     | 3.3405766     | when too big -> policy too random, should go down
# | policy_loss        | -0.011813248  |
# | serial_timesteps   | 100096        |
# | time_elapsed       | 656           |
# | total_timesteps    | 100096        |
# | value_loss         | 0.4478733     |
# --------------------------------------

# TODO add random search for hyperparameters

# TODO experiment with different nn sizes

# TODO make pyminijass work for easier experimentation -> ultra simple implementation: 1 suit 12 values distributed to 4 players

# TODO consider different random seeds!

# TODO make experiments with different reward functions

# TODO try Lstm policy

# TODO program evaluation of rl player strength!

# TODO try rules with trump


# TODO jassbot pypi package too!!!

import sys
import threading

print("Importing modules", file=sys.stderr)

import logging

from datetime import datetime

import numpy as np

from stable_baselines import PPO2
from stable_baselines.bench import load_results
from stable_baselines.results_plotter import ts2xy

from bot import rl as hp

from bot.rl import util
from bot.rl.jass_policy import JassPolicy
import os

print("Setting up experiment", file=sys.stderr)

os.environ["CUDA_VISIBLE_DEVICES"] = "-1"  # disable gpu usage because this apparently slows down training
logger = logging.getLogger(__name__)

# Create log dir
log_dir = "log/"
# this enables multiple experiments to be run in parallel --> there is one monitor file for each experiment
console_log_dir = log_dir + "console/" + hp.experiment_name + "/"
models_log_dir = log_dir + "models/"
tensorboard_log_dir = log_dir + "tensorboard/"
os.makedirs(log_dir, exist_ok=True)
os.makedirs(console_log_dir, exist_ok=True)
os.makedirs(models_log_dir, exist_ok=True)

env_name = 'Schieber-v0'
# time improvement is not very big with multiple cores
n_cpu = 1  # there are 8 cores on my local machine (MacBook Pro) and 16 on the server (Ubuntu 16.04)

env = util.init_env(env_name, console_log_dir, n_cpu, rules_reward=hp.rules_reward)

policy = "MlpLstmPolicy"
policy = "MlpLnLstmPolicy"
policy = "MlpPolicy"
policy = JassPolicy

model = PPO2(policy, env, n_steps=hp.n_steps, gamma=hp.gamma, nminibatches=hp.nminibatches,
             learning_rate=hp.learning_rate, verbose=1,
             tensorboard_log=tensorboard_log_dir)

model_name = "stich-higher-learning-rate_env=Schieber-v0_gamma=0.89_nsteps=90_learning_rate=0.001_policy=JassPolicy_model=PPO2_time=2019-01-11_10:16:41_final.pkl"
model = PPO2.load(models_log_dir + model_name, env=env)

model_name = model.__class__.__name__

time = datetime.now().strftime('%Y-%m-%d_%H:%M:%S')

log_name = f"{hp.experiment_name}_env={env_name}_gamma={hp.gamma}_nsteps={hp.n_steps}_learning_rate={hp.learning_rate}_policy={policy.__name__}_model={model_name}_time={time}"

training_log = open(f"{console_log_dir}/{log_name}.log", "a")
sys.stdout = training_log
logging.basicConfig(level=logging.INFO, filename=f"{console_log_dir}/{log_name}.log", datefmt='%H:%M:%S',
                    format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s')

best_mean_reward, n_steps = -np.inf, 0
model_file_name = f"{models_log_dir}{log_name}_best.pkl"


# open(model_file_name, 'a').close()  # create empty file here because otherwise saving does not work in the callback


def callback(_locals, _globals):
    """
    Callback called at each step (for DQN an others) or after n steps (see ACER or PPO2)
    :param _locals: (dict)
    :param _globals: (dict)
    """
    global n_steps, best_mean_reward, console_log_dir, model_file_name, log_name

    # This is invoked in every 10th update
    if (n_steps + 1) % 10 == 0:
        # Evaluate policy performance
        x, y = ts2xy(load_results(console_log_dir), 'timesteps')
        if len(x) > 0:
            mean_reward = np.mean(y[-100:])
            print(x[-1], 'timesteps', file=sys.stderr)
            print(
                "Best mean reward: {:.2f} - Last mean reward per episode: {:.2f}".format(best_mean_reward, mean_reward),
                file=sys.stderr)

            # New best model, you could save the agent here
            if mean_reward > best_mean_reward:
                best_mean_reward = mean_reward
                # Example for saving best model
                print("Saving new best model", file=sys.stderr)
                _locals['self'].save(model_file_name)
    n_steps += 1
    return True


start = datetime.now()

print("Learning model", file=sys.stderr)
model.learn(total_timesteps=hp.total_timesteps, tb_log_name=log_name, callback=callback)

training_time = datetime.now() - start
print(f"Training time: {training_time}", file=sys.stderr)

print("Saving final model", file=sys.stderr)
model.save(f"{models_log_dir}{log_name}_final.pkl")

print("Evaluating model", file=sys.stderr)
if n_cpu > 0:
    env = env.envs[0]  # we only need one environment for the evaluation
util.evaluate(env, model, n_games=10)

print(threading.enumerate(), file=sys.stderr)

env.close()
sys.exit()



from stable_baselines.common.policies import register_policy, MlpPolicy, FeedForwardPolicy


class JassPolicy(FeedForwardPolicy):
    def __init__(self, *args, **kwargs):
        super(JassPolicy, self).__init__(*args, **kwargs,
                                         net_arch=[dict(pi=[128, 128, 128], vf=[128, 128, 128])],
                                         feature_extraction="mlp")


# Register the policy, it will check that the name is not already taken
register_policy('JassPolicy', JassPolicy)



# ====================   Adjust this every time before conducting an experiment   ====================
experiment_name = "stich-higher-learning-rate"  # give a descriptive and memorable name here
# ====================================================================================================

rules_reward = False

if rules_reward:
    gamma = 0  # only care about immediate reward in the rule learning setting
else:
    gamma = 0.89  # care about rewards 9 steps into the future: 1 / (1 - gamma)

n_steps = 90  # = 10 games, determines the batch size
nminibatches = 10  # must be a divisor of n_steps!

learning_rate = 1e-3

total_timesteps = int(9e2)



# https://github.com/tomdee/president

from copy import deepcopy
import random
from bot.mcts import GameState, Card, ismcts


class KnockoutWhistState(GameState):
    """ A state of the game Knockout Whist.
        See http://www.pagat.com/whist/kowhist.html for a full description of the rules.
        For simplicity of implementation, this version of the game does not include the "dog's life" rule
        and the trump suit for each round is picked randomly rather than being chosen by one of the players.
    """

    def __init__(self, n):
        """ Initialise the game state. n is the number of players (from 2 to 7).
            """
        GameState.__init__(self)
        self.number_of_players = n
        self.player_to_move = 1
        self.tricks_in_round = 7
        self.player_hands = {p: [] for p in xrange(1, self.number_of_players + 1)}
        self.discards = []
        self.current_trick = []
        self.trump_suit = None
        self.tricks_taken = {}
        self.knocked_out = {p: False for p in
                           xrange(1, self.number_of_players + 1)}
        self._deal()

    def clone(self):
        """ Create a deep clone of this game state.
        """
        st = KnockoutWhistState(self.number_of_players)
        st.player_to_move = self.player_to_move
        st.tricks_in_round = self.tricks_in_round
        st.player_hands = deepcopy(self.player_hands)
        st.discards = deepcopy(self.discards)
        st.current_trick = deepcopy(self.current_trick)
        st.trump_suit = self.trump_suit
        st.tricks_taken = deepcopy(self.tricks_taken)
        st.knocked_out = deepcopy(self.knocked_out)
        return st

    def clone_and_randomize(self, observer):
        """ Create a deep clone of this game state, randomizing any information
        not visible to the specified observer player.
        """
        st = self.clone()

        # The observer can see his own hand and the cards in the current trick,
        # and can remember the cards played in previous tricks
        seen_cards = st.player_hands[observer] + st.discards + \
                     [card for (player, card) in st.current_trick]

        # The observer can't see the rest of the deck
        unseen_cards = [card for card in st._get_card_deck()
                        if card not in seen_cards]

        # _deal the unseen cards to the other players
        random.shuffle(unseen_cards)
        for p in xrange(1, st.number_of_players + 1):
            if p != observer:
                # _deal cards to player p
                # Store the size of player p's hand
                num_cards = len(self.player_hands[p])
                # Give player p the first num_cards unseen cards
                st.player_hands[p] = unseen_cards[:num_cards]
                # Remove those cards from unseen_cards
                unseen_cards = unseen_cards[num_cards:]

        return st

    @staticmethod
    def _get_card_deck():
        """ Construct a standard deck of 52 cards.
        """
        return [Card(rank, suit) for rank in xrange(2, 14 + 1) for suit in
                ['C', 'D', 'H', 'S']]

    def _deal(self):
        """ Reset the game state for the beginning of a new round, and _deal the cards.
        """
        self.discards = []
        self.current_trick = []
        self.tricks_taken = {p: 0 for p in xrange(1, self.number_of_players + 1)}

        # Construct a deck, shuffle it, and _deal it to the players
        deck = self._get_card_deck()
        random.shuffle(deck)
        for p in xrange(1, self.number_of_players + 1):
            self.player_hands[p] = deck[:self.tricks_in_round]
            deck = deck[self.tricks_in_round:]

        # Choose the trump suit for this round
        self.trump_suit = random.choice(['C', 'D', 'H', 'S'])

    def get_next_player(self, p):
        """ Return the player to the left of the specified player, skipping players who have been knocked out
        """
        next_player = (p % self.number_of_players) + 1
        # Skip any knocked-out players
        while next_player != p and self.knocked_out[next_player]:
            next_player = (next_player % self.number_of_players) + 1
        return next_player

    def do_move(self, move):
        """ update a state by carrying out the given move.
            Must update player_to_move.
        """
        # Store the played card in the current trick
        self.current_trick.append((self.player_to_move, move))

        # Remove the card from the player's hand
        self.player_hands[self.player_to_move].remove(move)

        # Find the next player
        self.player_to_move = self.get_next_player(self.player_to_move)

        # If the next player has already played in this trick, then the trick is over
        if any(True for (player, card) in self.current_trick if
               player == self.player_to_move):
            # Sort the plays in the trick: those that followed suit (in ascending rank order), then any trump plays (in ascending rank order)
            (leader, lead_card) = self.current_trick[0]
            suited_plays = [(player, card.rank) for (player, card) in
                           self.current_trick if card.suit == lead_card.suit]
            trump_plays = [(player, card.rank) for (player, card) in
                          self.current_trick if card.suit == self.trump_suit]
            sorted_plays = sorted(suited_plays,
                                 key=lambda (aplayer, rank): rank) + sorted(
                trump_plays, key=lambda (aplayer, rank): rank)
            # The winning play is the last element in sorted_plays
            trick_winner = sorted_plays[-1][0]

            # update the game state
            self.tricks_taken[trick_winner] += 1
            self.discards += [card for (player, card) in self.current_trick]
            self.current_trick = []
            self.player_to_move = trick_winner

            # If the next player's hand is empty, this round is over
            if not self.player_hands[self.player_to_move]:
                self.tricks_in_round -= 1
                self.knocked_out = {
                p: (self.knocked_out[p] or self.tricks_taken[p] == 0) for p in
                xrange(1, self.number_of_players + 1)}
                # If all but one players are now knocked out, the game is over
                if len([x for x in self.knocked_out.itervalues() if
                        x == False]) <= 1:
                    self.tricks_in_round = 0

                self._deal()

    def get_moves(self):
        """ Get all possible moves from this state.
        """
        hand = self.player_hands[self.player_to_move]
        if not self.current_trick:
            # May lead a trick with any card
            return hand
        else:
            (leader, lead_card) = self.current_trick[0]
            # Must follow suit if it is possible to do so
            cards_in_suit = [card for card in hand if card.suit == lead_card.suit]
            if cards_in_suit:
                return cards_in_suit
            else:
                # Can't follow suit, so can play any card
                return hand

    def get_result(self, player):
        """ Get the game result from the viewpoint of player.
        """
        return 0 if (self.knocked_out[player]) else 1

    def __repr__(self):
        """ Return a human-readable representation of the state
        """
        result = "Round %i" % self.tricks_in_round
        result += " | P%i: " % self.player_to_move
        result += ",".join(
            str(card) for card in self.player_hands[self.player_to_move])
        result += " | Tricks: %i" % self.tricks_taken[self.player_to_move]
        result += " | Trump: %s" % self.trump_suit
        result += " | Trick: ["
        result += ",".join(
            ("%i:%s" % (player, card)) for (player, card) in self.current_trick)
        result += "]"
        return result


def play_game():
    """ Play a sample game between two ismcts players.
    """
    state = KnockoutWhistState(4)

    while state.get_moves():
        print str(state)
        # Use different numbers of iterations (simulations, tree nodes) for different players
        if state.player_to_move == 1:
            m = ismcts(rootstate=state, itermax=1000, verbose=False)
        else:
            m = ismcts(rootstate=state, itermax=100, verbose=False)
        print "Best Move: " + str(m) + "\n"
        state.do_move(m)

    someone_won = False
    for p in xrange(1, state.number_of_players + 1):
        if state.get_result(p) > 0:
            print "Player " + str(p) + " wins!"
            someone_won = True
    if not someone_won:
        print "Nobody wins!"


if __name__ == "__main__":
    play_game()


import setuptools

setuptools.setup(
    name='jass_bot',
    version='0.0.3',
    install_requires=['gym>=0.10.8', 'gym-jass>=0.1.16', 'schieber>=0.1.8'],
    author="Joel Niklaus",
    author_email="me@joelniklaus.ch",
    description="An AI for the Schieber variant of the Swiss card game Jassen",
    license=open('LICENSE', "r").read(),
    long_description=open('README.md', "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/JoelNiklaus/JassBot",
    packages=setuptools.find_packages(exclude=('bot.rl.log', 'bot.cfr.tests', 'bot.mcts.tests', 'bot.rl.tests')),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)
